# Check the following urls for more info about Pan-STARRS:
#
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Image+Cutout+Service#PS1ImageCutoutService-ImportantFITSimageformat,WCS,andflux-scalingnotes
#     https://outerspace.stsci.edu/display/PANSTARRS/PS1+Stack+images#PS1Stackimages-Photometriccalibration
#
# For DES:
#
#     https://des.ncsa.illinois.edu/releases/dr1/dr1-docs/processing
#
# For SDSS:
#
#     https://www.sdss.org/dr12/algorithms/fluxcal/#SDSStoAB
#     https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
#
# Some parts of this notebook are based on https://github.com/djones1040/PS1_surface_brightness/blob/master/Surface%20Brightness%20Tutorial.ipynb and codes from Lluís Galbany

import os
import numpy as np

import sep
from astropy import wcs
from astropy.io import fits

from hostphot._constants import __workdir__
from hostphot.utils import (
    get_survey_filters,
    check_survey_validity,
    check_filters_validity,
    survey_zp,
    get_image_gain,
    get_image_readnoise,
    get_image_exptime,
    pixel2pixel,
    check_work_dir,
    uncertainty_calc,
)
from hostphot.objects_detect import extract_objects, plot_detected_objects
from hostphot.image_cleaning import remove_nan
from hostphot.dust import calc_extinction

import warnings
from astropy.utils.exceptions import AstropyWarning

sep.set_sub_object_limit(1e4)


# -------------------------------
def kron_flux(data, err, gain, objects, kronrad, scale):
    """Calculates the Kron flux.

    Parameters
    ----------
    data: 2D array
        Data of an image.
    err: float or 2D array
        Background error of the images.
    gain: float
        Gain value.
    objects: array
        Objects detected with `sep.extract()`.
    kronrad: float
        Kron radius.
    scale: float
        Scale of the Kron radius.

    Returns
    -------
    flux: array
        Kron flux.
    flux_err: array
        Kron flux error.
    """
    r_min = 1.75  # minimum diameter = 3.5

    if kronrad * np.sqrt(objects["a"] * objects["b"]) < r_min:
        print(f"Warning: using circular photometry")
        flux, flux_err, flag = sep.sum_circle(
            data,
            objects["x"],
            objects["y"],
            r_min,
            err=err,
            subpix=5,
            gain=1.0,
        )
    else:
        flux, flux_err, flag = sep.sum_ellipse(
            data,
            objects["x"],
            objects["y"],
            objects["a"],
            objects["b"],
            objects["theta"],
            scale * kronrad,
            err=err,
            subpix=5,
            gain=gain,
        )

    return flux, flux_err


def optimize_kron_flux(data, err, gain, objects, eps=0.0001):
    """Optimizes the Kron flux by iteration over different values.
    The stop condition is met when the change in flux is less that ``eps``.

    Parameters
    ----------
    data: ndarray
        Data of an image.
    err: float or ndarray
        Background error of the images.
    gain: float
        Gain value.
    objects: array
        Objects detected with :func:`sep.extract()`.
    eps: float, default ``0.0001`` (0.1%)
        Minimum percent change in flux allowed between iterations.

    Returns
    -------
    opt_flux: float
        Optimized Kron flux.
    opt_flux_err: float
        Optimized Kron flux error.
    opt_kronrad: float
        Optimized Kron radius.
    opt_scale: float
        Optimized scale of the Kron radius.
    """
    # iterate over kron radii
    for r in np.arange(1, 6.05, 0.05)[::-1]:
        kronrad, krflag = sep.kron_radius(
            data,
            objects["x"],
            objects["y"],
            objects["a"],
            objects["b"],
            objects["theta"],
            r,
        )
        opt_kronrad = kronrad
        if ~np.isnan(opt_kronrad):
            break

    if np.isnan(opt_kronrad):
        raise ValueError(
            "The Kron radius cannot be calculated. The image might have NaNs or the aperture is too close to the edge."
        )

    opt_flux = 0.0
    # iterate over scale
    scales = np.arange(1, 10, 0.01)
    for scale in scales:
        flux, flux_err = kron_flux(
            data, err, gain, objects, opt_kronrad, scale
        )
        flux, flux_err = flux[0], flux_err[0]

        calc_eps = np.abs(opt_flux - flux) / flux
        opt_flux, opt_flux_err = flux, flux_err
        opt_scale = scale
        if calc_eps < eps:
            opt_scale = scale
            break
        elif np.isnan(calc_eps):
            opt_scale = scales[-2]
            warnings.warn("Warning: the aperture might not fit in the image!")
            break

    return opt_flux, opt_flux_err, opt_kronrad, opt_scale


def extract_kronparams(
    name,
    host_ra,
    host_dec,
    filt,
    survey,
    ra=None,
    dec=None,
    bkg_sub=False,
    threshold=10,
    use_mask=True,
    optimize_kronrad=True,
    eps=0.0001,
    save_plots=True,
):
    """Calculates the aperture parameters for common aperture.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy declination of the galaxy in degrees.
    filt: str
        Filter to use to load the fits file.
    survey: str
        Survey to use for the zero-points and pixel scale.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: bool, default `False`
        If `True`, the image gets background subtracted.
    threshold: float, default `10`
        Threshold used by `sep.extract()` to extract objects.
    use_mask: bool, default `True`
        If `True`, the masked fits files are used. These must have
        been created beforehand.
    optimize_kronrad: bool, default `True`
        If `True`, the Kron radius is optimized, increasing the
        aperture size until the flux does not increase.
    eps: float, default ``0.0001`` (0.1%)
        Minimum percent change in flux allowed between iterations
        when optimizing the Kron radius.
    save_plots: bool, default `False`
        If `True`, the mask and galaxy aperture figures are saved.

    Returns
    -------
    gal_obj: array
        Galaxy object.
    img_wcs: WCS
        Image's WCS.
    kronrad: float
        Kron radius.
    scale: float
        Scale for the Kron radius.
    """
    check_survey_validity(survey)
    check_work_dir(__workdir__)
    obj_dir = os.path.join(__workdir__, name)
    if use_mask:
        suffix = "masked_"
    else:
        suffix = ""
    fits_file = os.path.join(obj_dir, f"{suffix}{survey}_{filt}.fits")

    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if bkg_sub:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    # extract objects
    gal_obj, nogal_objs = extract_objects(
        data_sub, bkg_rms, host_ra, host_dec, threshold, img_wcs
    )
    if optimize_kronrad:
        gain = 1  # doesn't matter here
        opt_res = optimize_kron_flux(data_sub, bkg_rms, gain, gal_obj, eps)
        flux, flux_err, kronrad, scale = opt_res
    else:
        scale = 2.5
        kronrad, krflag = sep.kron_radius(
            data_sub,
            gal_obj["x"],
            gal_obj["y"],
            gal_obj["a"],
            gal_obj["b"],
            gal_obj["theta"],
            6.0,
        )

    if save_plots:
        outfile = os.path.join(obj_dir, f"global_{survey}_{filt}.jpg")
        plot_detected_objects(
            data_sub, gal_obj, scale * kronrad, img_wcs, ra, dec, outfile
        )

    return gal_obj, img_wcs, kronrad, scale


def photometry(
    name,
    host_ra,
    host_dec,
    filt,
    survey,
    ra=None,
    dec=None,
    bkg_sub=False,
    threshold=10,
    use_mask=True,
    aperture_params=None,
    optimize_kronrad=True,
    eps=0.0001,
    save_plots=True,
):
    """Calculates the global aperture photometry of a galaxy using
    the Kron flux.

    **Note:** the galaxy must be ideally centred in the image.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy declination of the galaxy in degrees.
    filt: str
        Filter to use to load the fits file.
    survey: str
        Survey to use for the zero-points and pixel scale.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: bool, default `False`
        If `True`, the image gets background subtracted.
    threshold: float, default `10`
        Threshold used by `sep.extract()` to extract objects.
    use_mask: bool, default `True`
        If `True`, the masked fits files are used. These must have
        been created beforehand.
    aperture_params: tuple, default `None`
        Tuple with objects info and Kron parameters. Used for
        common aperture. If given, the Kron parameters are not
        re-calculated
    optimize_kronrad: bool, default `True`
        If `True`, the Kron radius is optimized, increasing the
        aperture size until the flux does not increase.
    eps: float, default ``0.0001`` (0.1%)
        Minimum percent change in flux allowed between iterations
        when optimizing the Kron radius.
    save_plots: bool, default `True`
        If `True`, the mask and galaxy aperture figures are saved.

    Returns
    -------
    mag: float
        Aperture magnitude.
    mag_err: float
        Error on the aperture magnitude.
    """
    check_survey_validity(survey)
    check_work_dir(__workdir__)
    obj_dir = os.path.join(__workdir__, name)
    if use_mask:
        suffix = "masked_"
    else:
        suffix = ""
    fits_file = os.path.join(obj_dir, f"{suffix}{survey}_{filt}.fits")

    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data

    exptime = get_image_exptime(header, survey)
    gain = get_image_gain(header, survey)
    readnoise = get_image_readnoise(header, survey)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if bkg_sub:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    if aperture_params is not None:
        gal_obj, img_wcs0, kronrad, scale = aperture_params

        gal_obj["x"], gal_obj["y"] = pixel2pixel(
            gal_obj["x"], gal_obj["y"], img_wcs0, img_wcs
        )

        flux, flux_err = kron_flux(
            data_sub, bkg_rms, gain, gal_obj, kronrad, scale
        )
        flux, flux_err = flux[0], flux_err[0]
    else:
        # extract objects
        gal_obj, nogal_objs = extract_objects(
            data_sub, bkg_rms, host_ra, host_dec, threshold, img_wcs
        )

        # aperture photometry
        # This uses what would be the default SExtractor parameters.
        # See https://sep.readthedocs.io/en/v1.1.x/apertures.html
        if optimize_kronrad:
            opt_res = optimize_kron_flux(data_sub, bkg_rms, gain, gal_obj, eps)
            flux, flux_err, kronrad, scale = opt_res
        else:
            kronrad, krflag = sep.kron_radius(
                data_sub,
                gal_obj["x"],
                gal_obj["y"],
                gal_obj["a"],
                gal_obj["b"],
                gal_obj["theta"],
                6.0,
            )
            scale = 2.5
            flux, flux_err = kron_flux(
                data_sub, bkg_rms, gain, gal_obj, kronrad, scale
            )
            flux, flux_err = flux[0], flux_err[0]

    zp_dict = survey_zp(survey)
    if zp_dict == "header":
        zp = header["MAGZP"]
    else:
        zp = zp_dict[filt]
    if survey == "PS1":
        zp += 2.5 * np.log10(exptime)

    mag = -2.5 * np.log10(flux) + zp
    mag_err = 2.5 / np.log(10) * flux_err / flux

    if survey == "WISE":
        # see Table 5 of
        # https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4c.html#wpro
        # apcor_dict = {
        #     "W1": -0.034,
        #     "W2": -0.041,
        #     "W3": 0.030,
        #     "W4": -0.029,
        # }  # in mags

        # correction assumed to be 0 mags as PSF fitting is not used.
        apcor_dict = {
            "W1": 0.0,
            "W2": 0.0,
            "W3": 0.0,
            "W4": 0.0,
        }  # in mags
        m_apcor = apcor_dict[filt]
        mag += m_apcor
        mag_err = 0.0  # flux_err already propagated below for this survey

    # correct extinction
    A_ext = calc_extinction(filt, survey, host_ra, host_dec)
    mag -= A_ext

    # error budget
    ap_area = np.pi * gal_obj["a"][0] * gal_obj["b"][0]
    extra_err = uncertainty_calc(
        flux,
        flux_err,
        survey,
        filt,
        ap_area,
        readnoise,
        gain,
        exptime,
        bkg_rms,
    )
    mag_err = np.sqrt(mag_err**2 + extra_err**2)

    if survey == "WISE":
        zp_unc = header["MAGZPUNC"]
        mag_err = np.sqrt(mag_err**2 + zp_unc**2)

    if save_plots:
        outfile = os.path.join(obj_dir, f"global_{survey}_{filt}.jpg")
        plot_detected_objects(
            data_sub, gal_obj, scale * kronrad, img_wcs, ra, dec, outfile
        )

    return mag, mag_err


def multi_band_phot(
    name,
    host_ra,
    host_dec,
    filters=None,
    survey="PS1",
    ra=None,
    dec=None,
    bkg_sub=False,
    threshold=10,
    use_mask=True,
    common_aperture=True,
    coadd_filters="riz",
    optimize_kronrad=True,
    eps=0.0001,
    save_plots=True,
):
    """Calculates multi-band aperture photometry of the host galaxy
    for an object.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy declination of the galaxy in degrees.
    filters: str, default, ``None``
        Filters to use to load the fits files. If `None` use all
        the filters of the given survey.
    survey: str, default ``PS1``
        Survey to use for the zero-points and pixel scale.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    bkg_sub: bool, default ``False``
        If `True`, the image gets background subtracted.
    threshold: float, default ``10``
        Threshold used by :func:`sep.extract()` to extract objects.
    use_mask: bool, default ``True``
        If ``True``, the masked fits files are used. These must have
        been created beforehand.
    common_aperture: bool, default ``True``
        If ``True``, use a coadd image for common aperture photometry.
    coadd_filters: str, default ``riz``
        Filters of the coadd image. Used for common aperture photometry.
    optimize_kronrad: bool, default ``True``
        If ``True``, the Kron radius is optimized, increasing the
        aperture size until the flux does not increase.
    eps: float, default ``0.0001`` (0.1%)
        Minimum percent change in flux allowed between iterations
        when optimizing the Kron radius.
    save_plots: bool, default ``True``
        If ``True``, the mask and galaxy aperture figures are saved.

    Returns
    -------
    results_dict: dict
        Dictionary with the object's photometry and other info.

    Examples
    --------
    >>> import hostphot.global_photometry as gp
    >>> name = 'SN2004eo'
    >>> host_ra, host_dec = 308.2092, 9.92755  # coords of host galaxy of SN2004eo
    >>> ra, dec =  308.22579, 9.92853 # coords of SN2004eo
    >>> results = gp.multi_band_phot(name, host_ra, host_dec,
                            survey=survey, ra=ra, dec=dec,
                            use_mask=True, common_aperture=True,
                            coadd_filters='riz', save_plots=True)
    """
    check_survey_validity(survey)
    if filters is None:
        filters = get_survey_filters(survey)
    else:
        check_filters_validity(filters, survey)

    results_dict = {
        "name": name,
        "host_ra": host_ra,
        "host_dec": host_dec,
        "survey": survey,
    }

    if common_aperture:
        aperture_params = extract_kronparams(
            name,
            host_ra,
            host_dec,
            coadd_filters,
            survey,
            ra,
            dec,
            bkg_sub,
            threshold,
            use_mask,
            optimize_kronrad,
            eps,
            save_plots,
        )
    else:
        aperture_params = None

    for filt in filters:
        mag, mag_err = photometry(
            name,
            host_ra,
            host_dec,
            filt,
            survey,
            ra,
            dec,
            bkg_sub,
            threshold,
            use_mask,
            aperture_params,
            optimize_kronrad,
            eps,
            save_plots,
        )
        results_dict[filt] = mag
        results_dict[f"{filt}_err"] = mag_err

    return results_dict



import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

import sep
from astropy.io import fits
from astropy import wcs, units as u
from astropy.coordinates import SkyCoord, concatenate
from astropy.convolution import (
    Gaussian2DKernel,
    convolve_fft,
    interpolate_replace_nans,
)

from hostphot._constants import __workdir__
from hostphot.image_cleaning import remove_nan
from hostphot.objects_detect import (
    extract_objects,
    find_gaia_objects,
    find_catalog_objects,
    cross_match,
)
from hostphot.utils import (
    check_survey_validity,
    pixel2pixel,
    update_axislabels,
    survey_pixel_scale,
)

import warnings
from astropy.utils.exceptions import AstropyWarning


# ----------------------------------------
def mask_image(data, objects, r=5, sigma=20):
    """Masks objects in an image (2D array) by convolving it with
    a 2D Gaussian kernel.

    Parameters
    ----------
    data: 2D array
        Image data.
    objects: array
        Objects extracted with :func:`sep.extract()`.
    r: float, default ``5``
        Scale of the semi-mayor and semi-minor axes
        of the ellipse of the `obejcts`.
    sigma: float, default ``20``
        Standard deviation in pixel units of the 2D Gaussian kernel
        used to convolve the image.

    Returns
    -------
    masked_data: 2D array
        Masked image data.
    """
    mask = np.zeros(data.shape, dtype=bool)
    sep.mask_ellipse(
        mask,
        objects["x"],
        objects["y"],
        objects["a"],
        objects["b"],
        objects["theta"],
        r=r,
    )

    masked_data = data.copy()
    masked_data[mask] = np.nan
    # mask data by convolving it with a 2D Gaussian kernel
    # with the same sigma in x and y
    kernel = Gaussian2DKernel(sigma)
    masked_data = interpolate_replace_nans(masked_data, kernel, convolve_fft)

    return masked_data


def create_mask(
    name,
    host_ra,
    host_dec,
    filt,
    survey,
    ra=None,
    dec=None,
    bkg_sub=False,
    threshold=15,
    sigma=8,
    crossmatch=False,
    extract_params=False,
    common_params=None,
    save_plots=True,
):
    """Calculates the aperture parameters for common aperture.

    Parameters
    ----------
    name: str
        Name of the object to find the path of the fits file.
    host_ra: float
        Host-galaxy right ascension in degrees.
    host_dec: float
        Host-galaxy declination in degrees.
    filt: str
        Filter to use to load the fits file.
    survey: str
        Survey to use for the zero-points and correct filter path.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting.
    bkg_sub: bool, default ``False``
        If ``True``, the image gets background subtracted.
    threshold: float, default `15`
        Threshold used by :func:`sep.extract()` to extract objects.
    sigma: float, default ``8``
        Standard deviation in pixel units of the 2D Gaussian kernel
        used to convolve the image.
    crossmatch: bool, default ``False``
        If ``True``, the detected objects are cross-matched with a
        Gaia catalog.
    extract_params: bool, default ``False``
        If ``True``, returns the parameters listed below.
    common_params: tuple, default ``None``
        Parameters to use for common masking of different filters.
        This are the same as the outputs of this function.
    save_plots: bool, default ``True``
        If ``True``, the mask and galaxy aperture figures are saved.

    Returns
    -------
    **This are only returned if ``extract_params==True``.**
    gal_obj: array
        Galaxy object.
    gal_obj: array
        Non-galaxy objects.
    img_wcs: WCS
        Image's WCS.
    pixel_scale: float
        Pixel scale for the survey.
    """
    check_survey_validity(survey)

    obj_dir = os.path.join(__workdir__, name)
    fits_file = os.path.join(obj_dir, f"{survey}_{filt}.fits")
    img = fits.open(fits_file)
    img = remove_nan(img)

    header = img[0].header
    data = img[0].data
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    data = data.astype(np.float64)
    bkg = sep.Background(data)
    bkg_rms = bkg.globalrms
    if bkg_sub:
        data_sub = np.copy(data - bkg)
    else:
        data_sub = np.copy(data)

    pixel_scale = survey_pixel_scale(survey)

    if common_params is None:
        # extract objects
        gal_obj, nogal_objs = extract_objects(
            data_sub, bkg_rms, host_ra, host_dec, threshold, img_wcs
        )
        # preprocessing: cross-match extracted objects with a catalog
        # using two Gaia catalogs as they do not always include the
        # same objects.
        if crossmatch:
            cat_coord1 = find_gaia_objects(host_ra, host_dec, img_wcs)
            cat_coord2 = find_catalog_objects(host_ra, host_dec, img_wcs)
            cat_coord = concatenate([cat_coord1, cat_coord2])
            nogal_objs = cross_match(nogal_objs, img_wcs, cat_coord)
    else:
        # use objects previously extracted
        # the pixels coordinates are updated accordingly
        gal_obj, nogal_objs, img_wcs0, pixel_scale0 = common_params
        # make copies to prevent altering the initial values
        gal_obj = gal_obj.copy()
        nogal_objs = nogal_objs.copy()
        scale = pixel_scale0 / pixel_scale

        gal_obj["x"], gal_obj["y"] = pixel2pixel(
            gal_obj["x"], gal_obj["y"], img_wcs0, img_wcs
        )
        gal_obj["a"] *= scale
        gal_obj["b"] *= scale

        nogal_objs["x"], nogal_objs["y"] = pixel2pixel(
            nogal_objs["x"], nogal_objs["y"], img_wcs0, img_wcs
        )
        nogal_objs["a"] *= scale
        nogal_objs["b"] *= scale

    masked_data = mask_image(data_sub, nogal_objs, sigma=sigma)
    img[0].data = masked_data
    outfile = os.path.join(obj_dir, f"masked_{survey}_{filt}.fits")
    img.writeto(outfile, overwrite=True)

    if save_plots:
        outfile = os.path.join(obj_dir, f"masked_{survey}_{filt}.jpg")
        plot_masked_image(
            data_sub, masked_data, nogal_objs, img_wcs, ra, dec, outfile
        )

    if extract_params:
        return gal_obj, nogal_objs, img_wcs, pixel_scale


def plot_masked_image(
    data, masked_data, objects, img_wcs, ra=None, dec=None, outfile=None
):
    """Plots the masked image together with the original image and
    the detected objects.

    Parameters
    ----------
    data: 2D array
        Image data.
    masked_data: 2D array
         Masked image data.
    objects: array
        Objects extracted with :func:`sep.extract()`.
    img_wcs: WCS
        Image's WCS.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """
    r = 4  # scale
    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(24, 10))
    ax0 = plt.subplot(131, projection=img_wcs)
    ax1 = plt.subplot(132, projection=img_wcs)
    ax2 = plt.subplot(133, projection=img_wcs)
    axes = [ax0, ax1, ax2]

    overlays = []
    for ax in axes:
        update_axislabels(ax)
        ax.imshow(
            data,
            interpolation="nearest",
            cmap="gray",
            vmin=m - s,
            vmax=m + s,
            origin="lower",
        )
    for ax in axes[1:]:
        ax.coords[1].set_axislabel("")
        ax.coords[1].set_ticklabel_visible(False)

    for i in range(len(objects)):
        e = Ellipse(
            xy=(objects["x"][i], objects["y"][i]),
            width=2 * r * objects["a"][i],
            height=2 * r * objects["b"][i],
            angle=objects["theta"][i] * 180.0 / np.pi,
        )
        e.set_facecolor("none")
        e.set_edgecolor("red")
        e.set_linewidth(1.5)
        ax1.add_artist(e)

    ax2.imshow(
        masked_data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    ax0.set_title("Initial Image", fontsize=24)
    ax1.set_title("Detected Objects", fontsize=24)
    ax2.set_title("Masked Image", fontsize=24)

    if (ra is not None) and (dec is not None):
        coord = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        x, y = img_wcs.world_to_pixel(coord)
        for ax in axes:
            ax.scatter(x, y, marker="*", s=100, c="g", edgecolor="gold")

    if outfile:
        plt.tight_layout()
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

import sep
from astroquery.gaia import Gaia
from astroquery.mast import Catalogs
from astropy import units as u
from astropy.coordinates import SkyCoord

from hostphot.utils import update_axislabels


def extract_objects(data, err, host_ra, host_dec, threshold, img_wcs):
    """Extracts objects and their ellipse parameters. The function :func:`sep.extract()`
    is used.

    Parameters
    ----------
    data: 2D array
        Image data.
    err: 2D array
        Background error of the image.
    host_ra: float
        Host-galaxy Right ascension of the galaxy in degrees.
    host_dec: float
        Host-galaxy Declination of the galaxy in degrees.
    threshold: float
        Source with flux above ``threshold*bkg_rms`` are extracted.
        See :func:`sep.extract()` for more information.

    Returns
    -------
    gal_obj: numpy array
        Galaxy object extracted.
    nogal_objs: numpy array
        All objects extracted except for the galaxy.
    """
    # extract objects with Source Extractor
    objects = sep.extract(data, threshold, err=err)

    gal_coords = SkyCoord(
        ra=host_ra * u.degree, dec=host_dec * u.degree, frame="icrs"
    )
    gal_x, gal_y = img_wcs.world_to_pixel(gal_coords)

    # find the galaxy
    x_diff = np.abs(objects["x"] - gal_x)
    y_diff = np.abs(objects["y"] - gal_y)
    dist = np.sqrt(x_diff**2 + y_diff**2)
    gal_id = np.argmin(dist)
    gal_obj = objects[gal_id : gal_id + 1]

    objs_id = [i for i in range(len(objects)) if i != gal_id]
    nogal_objs = objects.take(objs_id)

    return gal_obj, nogal_objs


def find_gaia_objects(ra, dec, img_wcs, rad=0.15):
    """Finds objects using the Gaia DR3 catalog for the given
    coordinates in a given radius.

    Parameters
    ----------
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    img_wcs: WCS object
        WCS of an image.
    rad: float, default ``0.15``
        Search radius in degrees.

    Returns
    -------
    gaia_coord: SkyCoor object
        Coordinates of the objects found.
    """
    Gaia.MAIN_GAIA_TABLE = "gaiaedr3.gaia_source"
    Gaia.ROW_LIMIT = -1
    coord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    width = u.Quantity(rad, u.deg)
    height = u.Quantity(rad, u.deg)
    try:
        gaia_cat = Gaia.query_object_async(
            coordinate=coord, width=width, height=height
        )
    except:
        print("No objects found with Gaia DR3, switching to DR2")
        Gaia.MAIN_GAIA_TABLE = "gaiadr2.gaia_source"
        gaia_cat = Gaia.query_object_async(
            coordinate=coord, width=width, height=height
        )

    gaia_ra = np.array(gaia_cat["ra"].value)
    gaia_dec = np.array(gaia_cat["dec"].value)
    gaia_coord = SkyCoord(
        ra=gaia_ra, dec=gaia_dec, unit=(u.degree, u.degree), frame="icrs"
    )

    return gaia_coord


def find_catalog_objects(ra, dec, img_wcs, rad=0.15):
    """Finds objects using the TESS image cutouts (Tic) catalog
    for the given coordinates in a given radius.

    **Note:** this catalog includes objects from these sources:
        HIP, TYC, UCAC, TWOMASS, SDSS, ALLWISE, GAIA, APASS, KIC

    Parameters
    ----------
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    img_wcs: WCS object
        WCS of an image.
    rad: float, default ``0.15``
        Search radius in degrees.

    Returns
    -------
    cat_coord: SkyCoor object
        Coordinates of the objects found.
    """
    coord = SkyCoord(ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs")
    cat_data = Catalogs.query_criteria(
        catalog="Tic", radius=rad, coordinates=coord, objType="STAR"
    )

    cat_ra = np.array(cat_data["ra"].value)
    cat_dec = np.array(cat_data["dec"].value)
    cat_coord = SkyCoord(
        ra=cat_ra, dec=cat_dec, unit=(u.degree, u.degree), frame="icrs"
    )

    return cat_coord


def cross_match(objects, img_wcs, coord, dist_thresh=1.0):
    """Cross-matches objects with a given set of coordinates.
    Those with a distance of less than ``dist_thresh`` are selected.

    Parameters
    ----------
    objects: array
        Objects detected with :func:`sep.extract()`.
    img_wcs: WCS
        WCS of the image from which the objects where extracted.
    coord: SkyCoor object
        Coordinates for the cross-match.
    dist_thresh: float, default ``1.0``
        Distance in arcsec to crossmatch the objects with
        the given coordinates.
    """
    # coordinates in arcsec
    objs_coord = img_wcs.pixel_to_world(objects["x"], objects["y"])
    objs_ra = objs_coord.ra.to(u.arcsec).value
    objs_dec = objs_coord.dec.to(u.arcsec).value

    cat_ra = np.array(coord.ra.to(u.arcsec).value)
    cat_dec = np.array(coord.dec.to(u.arcsec).value)

    objs_id = []
    for i in range(len(objects)):
        ra_dist = objs_ra[i] - cat_ra
        dec_dist = objs_dec[i] - cat_dec
        dist_arcsec = np.sqrt(ra_dist**2 + dec_dist**2)
        if any(dist_arcsec <= dist_thresh):
            objs_id.append(i)

    objs = objects.take(objs_id)

    return objs


def plot_detected_objects(
    data, objects, scale, img_wcs, ra=None, dec=None, outfile=None
):
    """Plots the objects extracted with :func:`sep.extract()``.

    Parameters
    ----------
    data: 2D array
        Data of an image.
    objects: array
        Objects detected with :func:`sep.extract()`.
    scale: float
        Scale of the ellipse's semi-mayor and semi-minor axes.
    img_wcs: WCS
        Image's WCS.
    ra: float, default ``None``
       Right ascension of an object, in degrees. Used for plotting the position of the object.
    dec: float, default ``None``
       Declination of an object, in degrees. Used for plotting the position of the object.
    outfile: str, default ``None``
        If given, path where to save the output figure.
    """
    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(10, 10))
    ax = plt.subplot(projection=img_wcs)
    update_axislabels(ax)

    im = ax.imshow(
        data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    e = Ellipse(
        xy=(objects["x"][0], objects["y"][0]),
        width=scale * objects["a"][0],
        height=scale * objects["b"][0],
        angle=objects["theta"][0] * 180.0 / np.pi,
    )

    e.set_facecolor("none")
    e.set_edgecolor("red")
    e.set_linewidth(1.5)
    ax.add_artist(e)

    if (ra is not None) and (dec is not None):
        coord = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        px, py = img_wcs.world_to_pixel(coord)
        ax.scatter(px, py, marker="*", s=140, c="g", edgecolor="gold")

    if outfile:
        plt.tight_layout()
        plt.savefig(outfile, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()



# -*- coding: utf-8 -*-

from . import utils
from . import cutouts
from . import rgb_images

from . import objects_detect
from . import image_masking
from . import image_cleaning
from . import objects_detect
from . import coadd
from . import dust

from . import local_photometry
from . import global_photometry
from . import interactive_aperture
from ._version import __version__
from ._constants import __workdir__



import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from astropy import wcs
from astropy.io import fits
from astropy.stats import sigma_clipped_stats
from photutils.utils import calc_total_error

import warnings
from astropy.utils.exceptions import AstropyWarning

import hostphot

hostphot_path = hostphot.__path__[0]
config_file = os.path.join(hostphot_path, "filters", "config.txt")
config_df = pd.read_csv(config_file, delim_whitespace=True)


def calc_sky_unc(image, exptime):
    """Calculates the uncertainty of the image from the
    sky standard deviation, sigma-clipped STD.

    Parameters
    ----------
    image: ndarray
        Image in a 2D numpy array.
    exptime: float
        Exposure time of the image.

    Returns
    -------
    error: float
        Estimated error of the image.
    """
    mask = image != 0
    avg, sky, sky_std = sigma_clipped_stats(image[mask], sigma=3.0)
    error = calc_total_error(image, sky_std, exptime)

    return error


def pixel2pixel(x1, y1, img_wcs1, img_wcs2):
    """Convert the pixel coordinates from one image to another."""
    coord1 = img_wcs1.pixel_to_world(x1, y1)
    x2, y2 = img_wcs2.world_to_pixel(coord1)

    return x2, y2


def check_survey_validity(survey):
    """Check whether the given survey is whithin the valid
    options.

    Parameters
    ----------
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.
    """
    global config_df
    surveys = list(config_df.survey)
    assert survey in surveys, f"survey '{survey}' not" f" in {surveys}"


def get_survey_filters(survey):
    """Gets all the valid filters for the given survey.

    Parameters
    ----------
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.

    Returns
    -------
    filters: str
        Filters for the given survey.
    """
    check_survey_validity(survey)

    global config_df
    survey_df = config_df[config_df.survey == survey]
    filters = survey_df.filters.values[0]

    if "," in filters:
        filters = filters.split(",")

    return filters


def survey_zp(survey):
    """Returns the zero-point for a given survey.

    **Note:** for ``PS1``, an extra :math:`+2.5*np.log10(exptime)`
    needs to be added afterwards.

    Parameters
    ----------
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.

    Returns
    -------
    zp_dict: dict or str
        Zero-points for all the filters in the given survey. If the survey zero-point
        is different for each image, the string ``header`` is returned.
    """
    check_survey_validity(survey)
    filters = get_survey_filters(survey)

    global config_df
    survey_df = config_df[config_df.survey == survey]
    zps = survey_df.zp.values[0]

    if zps == "header":
        return zps

    if "," in zps:
        zps = zps.split(",")
        zp_dict = {filt: float(zp) for filt, zp in zip(filters, zps)}
    else:
        zp_dict = {filt: float(zps) for filt in filters}

    return zp_dict


def get_image_gain(header, survey):
    """Returns the gain from an image's header.

    **Note:** for ``SDSS`` this is assumed to be one
    as it should already be included.

    Parameters
    ----------
    header: fits header
        Header of an image.
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.

    Returns
    -------
    gain: float
        Gain value.
    """
    check_survey_validity(survey)
    if survey == "PS1":
        gain = header["HIERARCH CELL.GAIN"]
    elif survey == "DES":
        gain = header["GAIN"]
    elif survey == "SDSS":
        gain = 1.0
    elif survey == "2MASS":
        # the value comes from https://wise2.ipac.caltech.edu/staff/jarrett/2mass/3chan/noise/
        # but it is different from the value of
        # https://iopscience.iop.org/article/10.1086/498708/pdf (8 e ADU^-1)
        gain = 10.0
    else:
        gain = 1.0

    return gain


def get_image_readnoise(header, survey):
    """Returns the read noise from an image's header.
    All values are per-pixel values.

    **Note:** for ``SDSS`` this is assumed to be zero
    as it should already be included.

    Parameters
    ----------
    header: fits header
        Header of an image.
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.

    Returns
    -------
    readnoise: float
        Read noise value.
    """
    check_survey_validity(survey)
    if survey == "PS1":
        readnoise = header["HIERARCH CELL.READNOISE"]
    elif survey == "DES":
        # see https://arxiv.org/pdf/0810.3600.pdf
        readnoise = 7.0  # electrons per pixel
    elif survey == "SDSS":
        readnoise = 0.0
    elif survey == "2MASS":
        # https://iopscience.iop.org/article/10.1086/498708/pdf
        # 6 combined images
        readnoise = 4.5 * np.sqrt(6)  # not used
    else:
        readnoise = 0.0

    return readnoise


def get_image_exptime(header, survey):
    """Returns the exposure time from an image's header.

    Parameters
    ----------
    header: fits header
        Header of an image.
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.

    Returns
    -------
    exptime: float
        Exposure time in seconds.
    """
    check_survey_validity(survey)
    if survey in ["PS1", "DES", "GALEX"]:
        exptime = float(header["EXPTIME"])
    elif survey == "WISE":
        # see: https://wise2.ipac.caltech.edu/docs/release/allsky/
        # and https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec1_1.html
        if header["BAND"] in [1, 2]:
            exptime = 7.7
        elif header["BAND"] in [3, 4]:
            exptime = 8.8
    elif survey == "2MASS":
        # https://iopscience.iop.org/article/10.1086/498708/pdf
        exptime = 7.8
    else:
        exptime = 1.0

    return exptime


def uncertainty_calc(
    flux,
    flux_err,
    survey,
    filt=None,
    ap_area=0.0,
    readnoise=0.0,
    gain=1.0,
    exptime=0.0,
    bkg_rms=0.0,
):
    """Calculates the uncertainty propagation.

    Parameters
    ----------
    flux: float
        Aperture flux.
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.
    filt: str, default ``None``
        Survey-specific filter.
    ap_area: float, default ``0.0``
        Aperture area.
    readnoise: float, default ``0.0``
        Image readnoise.
    gain: float, default ``1.0``
        Image gain.
    exptime: float, default ``0.0``
        Image exposure time.
    bkg_rms: float, default ``0.0``
        Background noise.

    Returns
    -------
    mag_err: float
        Extra uncertainty in magnitudes.
    """
    mag_err = 0.0
    if survey in ["PS1", "DES"]:
        # 1.0857 = 2.5/ln(10)
        extra_err = (
            1.0857 * np.sqrt(ap_area * (readnoise**2) + flux / gain) / flux
        )
        mag_err = np.sqrt(mag_err**2 + extra_err**2)

    if survey=="DES":
        unc_dict = {'g':2.6e-3, 'r':2.9e-3, 'i':3.4e-3,
                    'z':2.5e-3, 'Y':4.5e-3}
        extra_err = unc_dict[filt]
        mag_err = np.sqrt(mag_err ** 2 + extra_err ** 2)

    if survey=="SDSS":
        # https://data.sdss.org/datamodel/files/BOSS_PHOTOOBJ/frames/RERUN/RUN/CAMCOL/frame.html
        gain_dict = {'u':2.17, 'g':4.05, 'r':4.895,
                     'i':4.885, 'z':5.155}
        # dark variance
        dv_dict = {'u': 12.6025, 'g': 1.96, 'r': 1.8225,
                     'i': 7.84, 'z': 1.21}
        gain = gain_dict[filt]
        dark_variance = dv_dict[filt]
        extra_err = (
                1.0857 * np.sqrt(dark_variance + flux / gain) / flux
        )
        mag_err = np.sqrt(mag_err ** 2 + extra_err ** 2)
        mag_err = np.sqrt(mag_err ** 2 + extra_err ** 2)

    elif survey == "GALEX":
        CPS = flux
        if filt == "FUV":
            uv_err = -2.5 * (
                np.log10(CPS)
                - np.log10(
                    CPS
                    + (CPS * exptime + (0.050 * CPS * exptime) ** 2) ** 0.5
                    / exptime
                )
            )
        elif filt == "NUV":
            uv_err = -2.5 * (
                np.log10(CPS)
                - np.log10(
                    CPS
                    + (CPS * exptime + (0.027 * CPS * exptime) ** 2) ** 0.5
                    / exptime
                )
            )
        mag_err = np.sqrt(mag_err**2 + uv_err**2)

    elif survey == "2MASS":
        # see: https://wise2.ipac.caltech.edu/staff/jarrett/2mass/3chan/noise/
        S = flux
        N_c = 6  # number of coadd pixels
        k_z = 1.7  # kernel smoothing factor
        n_f = ap_area  # number of frame pixels in the aperture; aprox. as aperture area
        n_c = 4 * n_f  # number of coadd pixels in the aperture
        sigma_c = bkg_rms  # coadd noise; assumed to be ~background noise

        SNR = S / np.sqrt(
            S / (gain * N_c)
            + n_c * (2 * k_z * sigma_c) ** 2
            + (n_c * 0.024 * sigma_c) ** 2
        )
        mag_err = 1.0857 / SNR

    elif survey == "WISE":
        # see: https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec2_3f.html
        # see Table 5 of
        # https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4c.html#wpro
        # apcor_dict = {
        #     "W1": -0.034,
        #     "W2": -0.041,
        #     "W3": 0.030,
        #     "W4": -0.029,
        # }  # in mags

        # correction assumed to be 0 mags as PSF fitting is not used.
        apcor_dict = {
            "W1": 0.0,
            "W2": 0.0,
            "W3": 0.0,
            "W4": 0.0,
        }  # in mags
        m_apcor = apcor_dict[filt]
        f_apcor = 10 ** (-0.4 * m_apcor)
        F_src = f_apcor * flux

        N_p = (
            ap_area  # effective number of noise-pixels characterizing the PRF
        )
        # ratio of input (detector) pixel scale to output (Atlas Image) pixel scale
        pixel_scale_ratios = {"W1": 2, "W2": 2, "W3": 2, "W4": 4}
        Sin_Sout = pixel_scale_ratios[filt]
        F_corr = N_p * Sin_Sout**2

        k = 1
        N_A = N_B = ap_area
        sigma_conf = flux_err  # assumed to be the ~error in the aperture sum
        sigma_src = np.sqrt(
            f_apcor**2
            * F_corr
            * (flux_err**2 + k * (N_A**2) / N_B * bkg_rms**2)
            + sigma_conf**2
        )

        mag_err = np.sqrt(1.179 * sigma_src**2 / F_src**2)

    return mag_err


def survey_pixel_scale(survey):
    """Returns the pixel scale for a given survey.

    Parameters
    ----------
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.

    Returns
    -------
    pixel_scale: float
        Pixel scale in units of arcsec/pixel.
    """
    check_survey_validity(survey)

    global config_df
    survey_df = config_df[config_df.survey == survey]
    pixel_scale = survey_df.pixel_scale.values[0]

    return pixel_scale


def check_filters_validity(filters, survey):
    """Check whether the given filters are whithin the valid
    options for the given survey.

    Parameters
    ----------
    filters: str
        Filters to use, e,g, ``griz``.
    survey: str
        Survey name: ``PS1``, ``DES``, ``SDSS``, ``GALEX``, ``WISE``, ``2MASS``.
    """
    if filters is not None:
        valid_filters = get_survey_filters(survey)

        for filt in filters:
            message = (
                f"filter '{filt}' is not a valid option for "
                f"'{survey}' survey ({valid_filters})"
            )
            assert filt in valid_filters, message


def extract_filters(filters, survey):
    """Extracts transmission functions.

    Parameters
    ----------
    filters: str
        Filters to extract.
    survey: str
        Survey of the filters.

    Returns
    -------
    filters_dict: dict
        Dictionary with transmission functions
        and their respective wavelengths.
    """
    check_survey_validity(survey)
    check_filters_validity(filters, survey)

    filters_dict = {filt: None for filt in filters}
    if "WISE" in survey:
        survey == "WISE"  # for unWISE to use the same filters as WISE
    filters_path = os.path.join(hostphot.__path__[0], "filters", survey)

    for filt in filters:
        filt_file = os.path.join(filters_path, f"{survey}_{filt}.dat")
        wave, transmission = np.loadtxt(filt_file).T

        filters_dict[filt] = {"wave": wave, "transmission": transmission}

    return filters_dict


def integrate_filter(
    spectrum_wave,
    spectrum_flux,
    filter_wave,
    filter_response,
    response_type="photon",
):
    """Calcultes the flux density of an SED given a filter response.

    Parameters
    ----------
    spectrum_wave : array
        Spectrum's wavelength range.
    spectrum_flux : array
        Spectrum's flux density distribution.
    filter_wave : array
        Filter's wavelength range.
    filter_response : array
        Filter's response function.
    response_type : str, default ``photon``
        Filter's response type. Either ``photon`` or ``energy``.

    Returns
    -------
    flux_filter : float
        Flux density.
    """
    if response_type == "energy":
        filter_response = filter_response.copy() / filter_wave

    interp_response = np.interp(
        spectrum_wave, filter_wave, filter_response, left=0.0, right=0.0
    )
    int1 = np.trapz(
        spectrum_flux * interp_response * spectrum_wave, spectrum_wave
    )
    int2 = np.trapz(filter_response * filter_wave, filter_wave)
    flux_filter = int1 / int2

    return flux_filter


def check_work_dir(wokrdir):
    """Checks if the working directory exists. If it
    does not, one is created.

    Parameters
    ----------
    wokrdir: str
        Working directory path.
    """
    if not os.path.isdir(wokrdir):
        os.mkdir(wokrdir)


def clean_dir(directory):
    """Removes the directory if it is empty.

    Parameters
    ----------
    directory: str
        Directory path.
    """
    try:
        os.rmdir(directory)
    except:
        pass


def plot_fits(fits_file):
    """Plots a fits file.

    Parameters
    ----------
    fits_file: str
        Path to fits file.
    """
    img = fits.open(fits_file)
    header = img[0].header
    data = img[0].data

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        img_wcs = wcs.WCS(header, naxis=2)

    m, s = np.nanmean(data), np.nanstd(data)

    fig = plt.figure(figsize=(10, 10))
    ax = plt.subplot(projection=img_wcs)
    update_axislabels(ax)

    im = ax.imshow(
        data,
        interpolation="nearest",
        cmap="gray",
        vmin=m - s,
        vmax=m + s,
        origin="lower",
    )

    plt.show()


def update_axislabels(ax):
    """Updates the labels and ticks of a plot.

    Parameters
    ----------
    ax: `.axes.SubplotBase`.
        The axis of a subplot.
    """
    for i in range(2):
        ax.coords[i].set_ticklabel(size=16)
        if i == 0:
            ax.coords[i].set_axislabel("RA (J2000)", fontsize=20)
        else:
            ax.coords[i].set_axislabel("Dec (J2000)", fontsize=20)
            # ax.coords[i].set_ticklabel(rotation=65)



import os
import copy
import tarfile
import requests
import numpy as np
import pandas as pd
import multiprocessing as mp

from astropy.io import fits
from astropy.table import Table
from astropy import wcs, units as u
from astropy.coordinates import SkyCoord

import pyvo  # 2MASS
from pyvo.dal import sia  # DES
from astroquery.skyview import SkyView  # other surveys
from astroquery.mast import Observations  # for GALEX EXPTIME

from reproject import reproject_interp, reproject_exact
from reproject.mosaicking import find_optimal_celestial_wcs

from hostphot._constants import __workdir__
from hostphot.utils import (
    get_survey_filters,
    clean_dir,
    check_work_dir,
    check_survey_validity,
    check_filters_validity,
    survey_pixel_scale,
)

import warnings
from astropy.utils.exceptions import AstropyWarning


# PS1
# ----------------------------------------
def query_ps1(ra, dec, size=3, filters=None):
    """Query ps1filenames.py service to get a list of images

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    table: astropy Table
        Astropy table with the results.
    """
    survey = "PS1"
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)

    pixel_scale = survey_pixel_scale(survey)
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    url = (
        f"{service}?ra={ra}&dec={dec}&size={size_pixels}&format=fits&"
        f"filters={filters}"
    )

    table = Table.read(url, format="ascii")

    return table


def get_PS1_urls(ra, dec, size=3, filters=None):
    """Get URLs for images obtained with :func:`query_ps1()`.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    url_list: list
        List of URLs for the fits images.
    """
    survey = "PS1"
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)

    pixel_scale = survey_pixel_scale(survey)
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    table = query_ps1(ra, dec, size=size, filters=filters)
    url = (
        "https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
        f"ra={ra}&dec={dec}&size={size_pixels}&format=fits"
    )

    # sort filters from blue to red
    flist = ["grizy".find(x) for x in table["filter"]]
    table = table[np.argsort(flist)]

    base_url = url + "&red="
    url_list = []
    for filename in table["filename"]:
        url_list.append(base_url + filename)

    return url_list


def get_PS1_images(ra, dec, size=3, filters=None):
    """Gets PS1 fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizy``.

    Returns
    -------
    fits_files: list
        List of fits images.
    """
    survey = "PS1"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    fits_url = get_PS1_urls(ra, dec, size, filters)

    hdu_list = []
    for url, filt in zip(fits_url, filters):
        hdu = fits.open(url)
        hdu_list.append(hdu)

    return hdu_list


# DES
# ----------------------------------------
def get_DES_urls(ra, dec, fov, filters="grizY"):
    """Obtains the URLs of the DES images+weights with the
    largest exposure times in the given filters.

    Parameters
    ----------
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    fov: float
        Field of view in degrees.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizY``.

    Returns
    -------
    url_list: list
        List of URLs with DES images.
    url_w_list: list
        List of URLs with DES images weights.
    """
    if filters is None:
        filters = get_survey_filters("DES")
    check_filters_validity(filters, "DES")

    des_access_url = "https://datalab.noirlab.edu/sia/des_dr1"
    svc = sia.SIAService(des_access_url)
    imgs_table = svc.search(
        (ra, dec), (fov / np.cos(dec * np.pi / 180), fov), verbosity=2
    )
    if len(imgs_table) == 0:
        print(("Warning: empty table returned for " f"ra={ra}, dec={dec}"))
        return None, None

    imgs_df = pd.DataFrame(imgs_table)

    cond_stack = imgs_df["proctype"] == "Stack"
    cond_image = imgs_df["prodtype"] == "image"
    cond_weight = imgs_df["prodtype"] == "weight"

    url_list = []
    url_w_list = []
    for filt in filters:
        cond_filt = imgs_df["obs_bandpass"] == filt

        selected_images = imgs_df[cond_filt & (cond_stack & cond_image)]
        selected_weights = imgs_df[cond_filt & (cond_stack & cond_weight)]

        if len(selected_images) > 0:
            # pick image with longest exposure time
            max_exptime = np.argmax(selected_images["exptime"])
            row = selected_images.iloc[max_exptime]
            url = row["access_url"]  # get the download URL

            max_exptime = np.argmax(selected_weights["exptime"])
            row_w = selected_weights.iloc[max_exptime]
            url_w = row_w["access_url"]
        else:
            url = url_w = None
        url_list.append(url)
        url_w_list.append(url_w)

    return url_list, url_w_list


def get_DES_images(ra, dec, size=3, filters=None):
    """Gets DES fits images for the given coordinates and
    filters.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``grizY``.

    Returns
    -------
    hdu_list: list
        List of fits images.
    """
    survey = "DES"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        fov = (size * u.arcmin).to(u.degree).value
    else:
        fov = size.to(u.degree).value

    url_list, url_w_list = get_DES_urls(ra, dec, fov, filters)

    if url_list is None:
        return None

    hdu_list = []
    for url, url_w in zip(url_list, url_w_list):
        # combine image+weights on a single fits file
        image_hdu = fits.open(url)
        weight_hdu = fits.open(url_w)
        hdu = fits.PrimaryHDU(image_hdu[0].data, header=image_hdu[0].header)
        hdu_err = fits.ImageHDU(
            weight_hdu[0].data, header=weight_hdu[0].header
        )
        hdu_sublist = fits.HDUList([hdu, hdu_err])
        hdu_list.append(hdu_sublist)

    return hdu_list


# SDSS
# ----------------------------------------
def get_SDSS_images(ra, dec, size=3, filters=None):
    """Downloads a set of SDSS fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``ugriz``.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "SDSS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    # SkyView calls the filters in a different way
    filters_dict = {
        "u": "SDSSu",
        "g": "SDSSg",
        "r": "SDSSr",
        "i": "SDSSi",
        "z": "SDSSz",
    }
    skyview_filters = [filters_dict[filt] for filt in filters]

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

        hdu_list = []
        for filt, skyview_filter in zip(filters, skyview_filters):
            hdu = SkyView.get_images(
                position=coords,
                coordinates="icrs",
                pixels=str(size_pixels),
                survey=skyview_filter,
                width=size_arcsec,
                height=size_arcsec,
            )

            hdu_list.append(hdu[0])

    return hdu_list


# GALEX
# ----------------------------------------
def get_used_image(header):
    """Obtains the name of the image downloaded by SkyView.

    Parameters
    ----------
    header: fits header
        Header of an image.

    Returns
    -------
    used_image: str
        Name of the image.
    """
    image_line = header["HISTORY"][-3]
    used_image = image_line.split("/")[-1].split("-")[0]

    return used_image


def get_exptime(used_image, obs_table, filt):
    """Obtains the exposure time for a GALEX image downloaded
    with SkyView.

    Parameters
    ----------
    used_image: str
        GALEX name of the image downloaded by SkyView
    obs_table: astropy.Table
        Table obtained with astroquery.Observations.
    filt: str
        GALEX filter: 'NUV' or 'FUV'.

    Returns
    -------
    texp: float
        Exposure time.
    """
    galex_df = obs_table.to_pandas()
    galex_df = galex_df[galex_df.filters == filt]

    target_df = galex_df[galex_df.target_name == used_image]
    texp = target_df.t_exptime.values[0]

    return texp


def get_GALEX_images(ra, dec, size=3, filters=None):
    """Downloads a set of GALEX fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``FUV, NUV``.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "GALEX"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    # SkyView calls the filters in a different way
    filters_dict = {"NUV": "GALEX Near UV", "FUV": "GALEX Far UV"}
    skyview_filters = [filters_dict[filt] for filt in filters]

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )
        obs_table = Observations.query_criteria(
            coordinates=coords, radius=size_arcsec, obs_collection=[survey]
        )

        skyview_hdu_list = SkyView.get_images(
            position=coords,
            coordinates="icrs",
            pixels=str(size_pixels),
            survey=skyview_filters,
            width=size_arcsec,
            height=size_arcsec,
        )

        hdu_list = []
        for filt, hdu in zip(filters, skyview_hdu_list):
            # add exposure time
            used_image = get_used_image(hdu[0].header)
            texp = get_exptime(used_image, obs_table, filt)
            if texp is None:
                return None
            hdu[0].header["EXPTIME"] = texp
            hdu[0].header["COMMENT"] = "EXPTIME added by HostPhot"

            hdu_list.append(hdu)

    return hdu_list


# WISE
# ----------------------------------------
def get_WISE_images(ra, dec, size=3, filters=None):
    """Downloads a set of WISE fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses all WISE filters.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "WISE"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

        skyview_fits = SkyView.get_images(
            position=coords,
            coordinates="icrs",
            pixels=str(size_pixels),
            survey="WISE 3.4",
            width=size_arcsec,
            height=size_arcsec,
        )
        header = skyview_fits[0][0].header

        coadd_id = get_used_image(header)
        coadd_id1 = coadd_id[:4]
        coadd_id2 = coadd_id1[:2]

        # for more info: https://irsa.ipac.caltech.edu/ibe/docs/wise/allwise/p3am_cdd/#sample_code
        base_url = (
            "http://irsa.ipac.caltech.edu/ibe/data/wise/allwise/p3am_cdd/"
        )
        coadd_url = os.path.join(coadd_id2, coadd_id1, coadd_id)
        params_url = f"center={ra},{dec}&size={size_arcsec.value}arcsec&gzip=0"  # center and size of the image

        hdu_list = []
        for filt in filters:
            i = filt[-1]
            band_url = f"{coadd_id}-w{i}-int-3.fits"
            url = os.path.join(
                base_url, coadd_url, band_url + "?" + params_url
            )
            hdu = fits.open(url)
            hdu_list.append(hdu)

    return hdu_list

def get_unWISE_images(ra, dec, size=3, filters=None, version="allwise"):
    """Downloads a set of unWISE fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses all WISE filters.
    version: str, default ``allwise``
        Version of the unWISE images. Either ``allwise``, ``neo1`` or ``neo2``.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = f"unWISE{version}"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if version in ["neo1", "neo2"]:
        # these only have W1 and W2 data
        if "W3" in filters:
            filters.remove("W3")
        if "W4" in filters:
            filters.remove("W4")

    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec)
    else:
        size_arcsec = size.to(u.arcsec)

    pixel_scale = survey_pixel_scale(survey)
    size_pixels = int(size_arcsec.value / pixel_scale)
    assert size_pixels <= 1024, "Maximum cutout size for unWISE is 1024 pixels"

    bands = ''.join(filt[-1] for filt in filters)  # e.g. 1234

    # for more info: http://unwise.me/imgsearch/
    base_url = "http://unwise.me/cutout_fits?"
    params_url = f"version={version}&ra={ra}&dec={dec}&size={size_pixels}&bands={bands}"
    master_url = os.path.join(base_url, params_url)

    response = requests.get(master_url, stream=True)
    target_file = 'unWISE_images.tar.gz'  # current directory
    if response.status_code == 200:
        with open(target_file, 'wb') as f:
            f.write(response.raw.read())

    hdu_list = []
    with tarfile.open(target_file) as tar_file:
        for filt in filters:
            fits_file = f'*{filt.lower()}-img-m.fits'
            tar_file.extract(fits_file, '.')
            hdu = fits.open(fits_file)
            hdu_list.append(hdu)
            os.remove(fits_file)

    os.remove(target_file)

    return hdu_list


# 2MASS
# ----------------------------------------
def get_2MASS_images(ra, dec, size=3, filters=None):
    """Downloads a set of 2MASS fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use. If ``None``, uses ``FUV, NUV``.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "2MASS"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_degree = (size * u.arcmin).to(u.degree)
    else:
        size_degree = size.to(u.degree)
    size_degree = size_degree.value

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

        twomass_services = pyvo.regsearch(
            servicetype="image", keywords=["2mass"]
        )
        table = twomass_services[0].search(pos=coords, size=size_degree)
        twomass_df = table.to_table().to_pandas()
        twomass_df = twomass_df[twomass_df.format == "image/fits"]

        # for more info: https://irsa.ipac.caltech.edu/ibe/docs/twomass/allsky/allsky/#main
        base_url = (
            "https://irsa.ipac.caltech.edu/ibe/data/twomass/allsky/allsky"
        )

        hdu_list = []
        for i, filt in enumerate(filters):
            band_df = twomass_df[twomass_df.band == filt]
            fname = band_df.download.values[0].split("=")[-1]
            hemisphere = band_df.hem.values[0]
            ordate = band_df.date.values[0]
            scanno = band_df.scan.values[0]

            tile_url = os.path.join(f"{ordate}{hemisphere}", f"s{scanno}")
            fits_url = os.path.join("image", f"{fname}.gz")
            params_url = f"center={ra},{dec}&size={size_degree}degree&gzip=0"  # center and size of the image

            url = os.path.join(base_url, tile_url, fits_url + "?" + params_url)
            hdu = fits.open(url)
            hdu_list.append(hdu)

    return hdu_list


# HST
# ----------------------------------------
def get_HST_images(ra, dec, size=3, filters=None):
    """Downloads a set of HST fits images for a given set
    of coordinates and filters using SkyView.

    Parameters
    ----------
    ra: str or float
        Right ascension in degrees.
    dec: str or float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters to use.

    Return
    ------
    hdu_list: list
        List with fits images for the given filters.
        `None` is returned if no image is found.
    """
    survey = "HST"
    if filters is None:
        filters = get_survey_filters(survey)
    check_filters_validity(filters, survey)

    if isinstance(size, (float, int)):
        size_degree = (size * u.arcmin).to(u.degree)
    else:
        size_degree = size.to(u.degree)
    size_degree = size_degree.value

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        coords = SkyCoord(
            ra=ra, dec=dec, unit=(u.degree, u.degree), frame="icrs"
        )

    obs_table = Observations.query_criteria(
        coordinates=coords, radius=size_degree, obs_collection=["HST"]
    )
    obs_df = obs_table.to_pandas()

    df_list = []
    for inst in obs_df.instrument_name.unique():
        inst_df = obs_df[obs_df.instrument_name == inst]

        for filt in inst_df.filters.unique():
            if not filt.startswith("F") or ";" in filt:
                continue  # skip these odd filters
            filt_df = inst_df[inst_df.filters == filt]

            dist = np.sqrt(
                (filt_df.s_dec.values - dec) ** 2
                + (filt_df.s_ra.values - ra) ** 2
            )
            filt_df = filt_df[
                dist == dist.min()
            ]  # take the most centred image
            filt_df = filt_df[:1]  # take first image
            df_list.append(filt_df)
    obs_df = pd.concat(df_list)

    # identify products to download
    data_products = Observations.get_product_list(Table.from_pandas(obs_df))
    dp_df = data_products.to_pandas()
    dp_df = dp_df[dp_df.productType == "SCIENCE"]

    df_list = []
    for filt in obs_df.filters.unique():
        if filt.lower() in dp_df.dataURI.values:
            filt_df = dp_df[dp_df.dataURI.contains(filt.lower())]
        elif filt.upper() in dp_df.dataURI.values:
            filt_df = dp_df[dp_df.dataURI.contains(filt.upper())]
        else:
            continue

        if "HLA" in filt_df.obs_collection:
            # choose images from the legacy archive
            filt_df = filt_df[filt_df.obs_collection == "HLA"]

        filt_df = filt_df[:1]  # take first image
        df_list.append(filt_df)
    dp_df = pd.concat(df_list)

    # download images
    Observations.download_products(
        Table.from_pandas(dp_df), productType="SCIENCE", extension="fits"
    )

    #return hdu_list


# Check orientation
# ------------------
def match_wcs(fits_files):
    """Matches the WCS of all the images, taking the first ones as
    a reference.

    Parameters
    ----------
    fits_files: list
        List of fits files.

    Returns
    -------
    matched_fits_files: list
        List of fits files with matched WCS.
    """
    matched_fits_files = copy.deepcopy(fits_files)
    # some hdu have data + error
    hdu0 = matched_fits_files[0][0]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", AstropyWarning)
        wcs0 = wcs.WCS(hdu0.header)

        for hdu in matched_fits_files[1:]:
            data, footprint = reproject_interp(hdu[0], hdu0.header)
            hdu[0].data = data
            hdu[0].header.update(wcs0.to_header())

    return matched_fits_files


# Master Function
# ----------------------------------------
def download_images(
    name, ra, dec, size=3, filters=None, overwrite=False, survey="PS1"
):
    """Download images for a given object in the given filters of a
    given survey.

    Parameters
    ----------
    name: str
        Name used for tracking the object in your local
        directory.
    ra: float
        Right ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str, default ``None``
        Filters for the images.
    overwrite: bool, default ``False``
        If ``True``, the images are overwritten if they already
        exist.
    survey: str, default ``PS1``
        Survey used to download the images

    Examples
    --------
    >>> from hostphot.cutouts import download_images
    >>> name = 'SN2004eo'
    >>> host_ra, host_dec = 308.2092, 9.92755  # coords of host galaxy of SN2004eo
    >>> survey = 'PS1'
    >>> download_images(name, host_ra, host_dec, survey=survey)
    """
    check_survey_validity(survey)
    check_filters_validity(filters, survey)
    if filters is None:
        filters = get_survey_filters(survey)

    check_work_dir(__workdir__)
    obj_dir = os.path.join(__workdir__, name)
    if not os.path.isdir(obj_dir):
        os.mkdir(obj_dir)

    # download fits files
    if survey == "PS1":
        hdu_list = get_PS1_images(ra, dec, size, filters)
    elif survey == "DES":
        hdu_list = get_DES_images(ra, dec, size, filters)
    elif survey == "SDSS":
        hdu_list = get_SDSS_images(ra, dec, size, filters)
    elif survey == "GALEX":
        hdu_list = get_GALEX_images(ra, dec, size, filters)
    elif "WISE" in survey:
        if survey == "WISE":
            hdu_list = get_WISE_images(ra, dec, size, filters)
        else:
            wise_version = survey.split("WISE")[-1]
            hdu_list = get_unWISE_images(ra, dec, size,
                                           filters, wise_version)
    elif survey == "2MASS":
        hdu_list = get_2MASS_images(ra, dec, size, filters)

    if hdu_list:
        # this corrects any possible shifts between the images
        # fits_files = match_wcs(fits_files)

        # fix wcs (some have rotated wcs)
        if survey in ["XXX"]:
            for hdu in hdu_list:
                fits_file = hdu[0]
                wcs_out, shape_out = find_optimal_celestial_wcs(
                    [fits_file], auto_rotate=True
                )
                fixed_data, footprint = reproject_exact(
                    fits_file, wcs_out, shape_out=shape_out
                )
                fits_file.header.update(wcs_out.to_header())
                fits_file.data = fixed_data

        for hdu, filt in zip(hdu_list, filters):
            outfile = os.path.join(obj_dir, f"{survey}_{filt}.fits")
            if not os.path.isfile(outfile):
                hdu.writeto(outfile)
            else:
                if overwrite:
                    hdu.writeto(outfile, overwrite=overwrite)
                else:
                    continue

    # remove directory if it ends up empty
    clean_dir(obj_dir)


def pool_download(
    df=None,
    name=None,
    ra=None,
    dec=None,
    size=600,
    filters=None,
    overwrite=False,
    survey="PS1",
    processes=8,
):
    """Downloads images for multiple objects using parallelisation.

    Parameters
    ----------
    df: DataFrame, default ``None``
        DataFrame with the values of the argmuents. If this is given,
        ``name``, ``ra`` and ``dec`` should be the names of the columns in
        ``df``.
    name: str or list-like, default ``None``
        Name used for tracking the object in your local
        directory.
    ra: float or list-like, default ``None``
        Right ascension in degrees of the center of the image.
    dec: float or list-like, default ``None``
        Declination in degrees of the center of the image.
    size: int, default ``600``
        Image size in pixels.
    filters: str, default ``None``
        Filters for the images. If ``None``, use all the available
        filters.
    overwrite: bool, default ``False``
        If ``True``, the images are overwritten if they already
        exist.
    survey: str, default ``PS1``
        Survey used to download the images
    processes: floar, default ``8``
        Number of processes to use for the parallelisation.
    """
    local_dict = locals()  # get dictionary of input arguments
    variable_args = ["name", "ra", "dec"]
    ignore_args = ["df", "processes"]
    args_dict = {
        key: local_dict[key]
        for key in local_dict.keys()
        if key not in ignore_args
    }

    # adapt the input to feed it to the parallelisation function
    input_dict = args_dict.copy()
    if df is None:
        length = len(name)
    else:
        length = len(df)
        for key in variable_args:
            var_name = args_dict[key]
            input_dict[key] = df[var_name].values

    for key in args_dict.keys():
        if key not in variable_args:
            input_dict[key] = [args_dict[key]] * length

    # transpose list
    input_args = list(map(list, zip(*input_dict.values())))

    pool = mp.Pool(processes)
    pool.starmap(download_images, (args for args in input_args))



import setuptools

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

with open("requirements.txt") as requirements_file:
    requirements = requirements_file.read().splitlines()

with open("src/hostphot/_version.py") as version_file:
    for line in version_file:
        if "__version__" in line:
            __version__ = line.split()[-1].replace('"', "")

setuptools.setup(
    name="HostPhot",
    version=__version__,
    author="Tomás Enrique Müller-Bravo and Lluís Galbany",
    author_email="t.e.muller-bravo@ice.csic.es",
    license="MIT",
    description="Global and local photometry of galaxies hosting supernovae or other transients",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/temuller/hostphot",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    # scripts=["bin/hostphot"],
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    package_data={"hostphot": ["filters/*"]},
    include_package_data=True,
)



# based on https://github.com/tremou/RGB_MSUastro

import os
import math
import requests
import numpy as np
import matplotlib.pyplot as plt
from astropy.io import fits
from astropy import units as u
from astropy.table import Table

from hostphot.utils import survey_pixel_scale


def linear(inputArray, scale_min=None, scale_max=None):
    """Performs linear scaling of the input numpy array.

    Parameters
    ----------
    inputArray: numpy array
            Image data array.
    scale_min: float, default ``None``
            Minimum data value.
    scale_max: float, default ``None``
            Maximum data value.

    Returns
    -------
    imageData: numpy array
            Scaled image data array.
    """
    imageData = np.array(inputArray, copy=True)

    if scale_min == None:
        scale_min = np.nanmin(imageData)
    if scale_max == None:
        scale_max = np.nanmax(imageData)

    imageData = imageData.clip(min=scale_min, max=scale_max)
    imageData = (imageData - scale_min) / (scale_max - scale_min)
    indices = np.where(imageData < 0)
    imageData[indices] = 0.0
    indices = np.where(imageData > 1)
    imageData[indices] = 1.0

    return imageData


def sqrt(inputArray, scale_min=None, scale_max=None):
    """Performs :func:`sqrt` scaling of the input numpy array.

    Parameters
    ----------
    inputArray: numpy array
            Image data array.
    scale_min: float, default ``None``
            Minimum data value.
    scale_max: float, default ``None``
            Maximum data value.

    Returns
    -------
    imageData: numpy array
            Scaled image data array.
    """
    imageData = np.array(inputArray, copy=True)

    if scale_min == None:
        scale_min = np.nanmin(imageData)
    if scale_max == None:
        scale_max = np.nanmax(imageData)

    imageData = imageData.clip(min=scale_min, max=scale_max)
    imageData = imageData - scale_min
    indices = np.where(imageData < 0)
    imageData[indices] = 0.0
    imageData = np.sqrt(imageData)
    imageData = imageData / math.sqrt(scale_max - scale_min)

    return imageData


def log(inputArray, scale_min=None, scale_max=None):
    """Performs :func:`log10` scaling of the input numpy array.

    Parameters
    ----------
    inputArray: numpy array
            Image data array.
    scale_min: float, default ``None``
            Minimum data value.
    scale_max: float, default ``None``
            Maximum data value.

    Returns
    -------
    imageData: numpy array
            Scaled image data array.
    """
    imageData = np.array(inputArray, copy=True)

    if scale_min == None:
        scale_min = np.nanmin(imageData)
    if scale_max == None:
        scale_max = np.nanmax(imageData)

    factor = math.log10(scale_max - scale_min)
    indices0 = np.where(imageData < scale_min)
    indices1 = np.where((imageData >= scale_min) & (imageData <= scale_max))
    indices2 = np.where(imageData > scale_max)
    imageData[indices0] = 0.0
    imageData[indices2] = 1.0
    try:
        imageData[indices1] = np.log10(imageData[indices1]) / factor
    except:
        print("Error on math.log10 for ", (imageData[indices1] - scale_min))

    return imageData


def asinh(inputArray, scale_min=None, scale_max=None, non_linear=2.0):
    """Performs :func:`asinh` scaling of the input numpy array.

    Parameters
    ----------
    inputArray: numpy array
            Image data array.
    scale_min: float, default ``None``
            Minimum data value.
    scale_max: float, default ``None``
            Maximum data value.
    non_linear: float, default ``2.0``
            Non-linearity factor.

    Returns
    -------
    imageData: numpy array
            Scaled image data array.
    """
    imageData = np.array(inputArray, copy=True)

    if scale_min == None:
        scale_min = np.nanmin(imageData)
    if scale_max == None:
        scale_max = np.nanmax(imageData)

    factor = np.arcsinh((scale_max - scale_min) / non_linear)
    indices0 = np.where(imageData < scale_min)
    indices1 = np.where((imageData >= scale_min) & (imageData <= scale_max))
    indices2 = np.where(imageData > scale_max)
    imageData[indices0] = 0.0
    imageData[indices2] = 1.0
    imageData[indices1] = (
        np.arcsinh((imageData[indices1] - scale_min) / non_linear) / factor
    )

    return imageData


def create_RGB_image(
    outfile=None,
    survey="PS1",
    filters="zir",
    images_dir="",
    scaling="linear",
    scaling_params=None,
):
    """Creates an RGB image.

    Parameters
    ----------
    outfile: str, default ``None``
        Output file name. If `None`, no output is saved.
    survey: str, default ``PS1``
        Survey used for the filters.
    filters: str, defaul ``zir``
        Filters used to create the RGB image (in that order). Must be three filters.
    images_dir: str, default `''`
        Directory where to find the images.
    scaling: str
        Type of scaling. Choose between ``linear``, ``sqrt``, ``log`` and ``asinh``.
    scaling_params: dict
        Dictionary with the parameters for the ``scaling`` function. If ``None``,
        use the default values.

    Examples
    --------
    >>> # Example of scaling_params:
    >>> scaling_params = {'R':{'min':r_med/100, 'max':r_med*30},
                                      'G':{'min':g_med/1000, 'max':g_med*150},
                                      'B':{'min':b_med/30, 'max':b_med*100}}
    """

    assert len(filters) == 3, "Three filters should be given."
    scalings = ["linear", "sqrt", "log", "asinh"]
    assert (
        scaling in scalings
    ), f"Not a valid scaling (choose between {scalings})"

    rgb_files = []
    for filt in filters:
        filt_image = os.path.join(images_dir, f"{survey}_{filt}.fits")
        rgb_files.append(filt_image)

    r_img = fits.getdata(rgb_files[0], ext=0)
    r_med = np.nanmedian(r_img)
    g_img = fits.getdata(rgb_files[1], ext=0)
    g_med = np.nanmedian(g_img)
    b_img = fits.getdata(rgb_files[2], ext=0)
    b_med = np.nanmedian(b_img)
    imgs = [r_img, g_img, b_img]

    if scaling_params is None:
        scaling_params = {
            "R": {"min": r_med / 100, "max": r_med * 30},
            "G": {"min": g_med / 1000, "max": g_med * 150},
            "B": {"min": b_med / 30, "max": b_med * 100},
        }
    scaling_func = eval(scaling)  # scaling function

    img = np.zeros((r_img.shape[0], r_img.shape[1], 3), dtype=float)
    for i, filt_img in enumerate(imgs):
        img_scalings = list(scaling_params.values())[i]
        scale_min, scale_max = img_scalings.values()
        img[:, :, i] = scaling_func(
            filt_img, scale_min=scale_min, scale_max=scale_max
        )

    fig = plt.figure(figsize=(12, 12))
    plt.clf()
    plt.imshow(img, origin="lower")
    plt.title(f"Blue = {filters[0]}, Green = {filters[1]}, Red = {filters[2]}")
    if outfile is not None:
        plt.savefig(outfile)


# ---------------------------------------


def get_PS1_url(ra, dec, size=3, filters="grizy", data_format="jpg"):
    """Get URL for the colour image.

    Parameters
    ----------
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str
        Filters to include.
    data_format: str
        Data format (options are ``jpg`` or ``png``).

    Returns
    -------
    url: str
        The image's URL for a colour image.
    """
    assert data_format in ["jpg", "png"], (
        "format must be one of " "'jpg' or 'png'"
    )
    assert len(filters) >= 3, "must choose at least 3 filters"

    pixel_scale = survey_pixel_scale("PS1")
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value
    size_pixels = int(size_arcsec / pixel_scale)

    # get table with images
    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    table_url = (
        f"{service}?ra={ra}&dec={dec}&size={size_pixels}&format=fits&"
        f"filters={filters}"
    )
    table = Table.read(table_url, format="ascii")

    # url for colour image
    url = (
        "https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
        f"ra={ra}&dec={dec}&size={size_pixels}&format={data_format}"
        f"&output_size={size_pixels}"
    )

    # sort filters from red to blue
    flist = ["yzirg".find(x) for x in table["filter"]]
    table = table[np.argsort(flist)]

    if len(table) > 3:
        # pick only 3 filters
        table = table[[0, len(table) // 2, len(table) - 1]]

    for i, param in enumerate(["red", "green", "blue"]):
        url = url + f"&{param}={table['filename'][i]}"

    return url


def get_PS1_RGB_image(outfile, ra, dec, size=3, filters="grizy"):
    """Downloads an RGB image from the PS1 server.

    Parameters
    ----------
    outfile: str
        URL of the image to be downloaded
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.
    filters: str
        Filters to include.

    Returns
    -------
    url: str
        The image's URL for a colour image.
    """
    data_format = os.path.splitext(outfile)[-1][1:]
    remote_url = get_PS1_url(ra, dec, size, filters, data_format)

    data = requests.get(remote_url)
    with open(outfile, "wb") as file:
        file.write(data.content)


def get_SDSS_RGB_image(outfile, ra, dec, size=3):
    """Downloads an RGB image from the SDSS server.

    Parameters
    ----------
    remote_url: str
        URL of the image to be downloaded
    ra: float
        Right Ascension in degrees.
    dec: float
        Declination in degrees.
    size: float or ~astropy.units.Quantity, default ``3``
        Image size. If a float is given, the units are assumed to be arcmin.

    Returns
    -------
    url: str
        The image's URL for a colour image.
    """
    pixel_scale = survey_pixel_scale("SDSS")
    if isinstance(size, (float, int)):
        size_arcsec = (size * u.arcmin).to(u.arcsec).value
    else:
        size_arcsec = size.to(u.arcsec).value

    size_pixels = int(size_arcsec / pixel_scale)

    remote_url = (
        "http://skyserver.sdss.org/dr16/SkyServerWS/ImgCutout/"
        f"getjpeg?ra={ra}&dec={dec}&scale={pixel_scale}"
        f"&width={size_pixels}&height={size_pixels}"
    )

    data = requests.get(remote_url)
    with open(outfile, "wb") as file:
        file.write(data.content)
